﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Primeval
{
    /// <summary>
    /// A combat bot used in the game
    /// </summary>
    class Bot : IComparable
    {
        #region Static
        public static List<Bot> bots;

        static Bot()
        {
            bots = new List<Bot>();
        }

        public static void Reset()
        {
            bots.Clear();
        }

        public static int BotCount { get { return bots.Count; } }

        public static void AddBot(Bot r)
        {
            bots.Add(r);
        }

        public static void UpdateAll(GameTime gameTime)
        {
            for (int c = bots.Count - 1; c >= 0; c--)
            {
                bots[c].Update(gameTime);
                if (bots[c].Die) bots.RemoveAt(c);
            }
        }

        public static void DrawAll(SpriteBatch spriteBatch)
        {
            foreach (Bot r in bots) r.Draw(spriteBatch);
        }

        public static void DrawAllUI(SpriteBatch spriteBatch)
        {
            foreach (Bot r in bots) r.DrawUI(spriteBatch);
        }
        #endregion

        public static Texture2D frame;

        public static void LoadContent()
        {
            frame = Game1.content.Load<Texture2D>("botbase");
        }

        /// <summary>
        /// The genome that built this bot
        /// </summary>
        private Genome genome;

        /// <summary>
        /// A list of weapons this bot possesses
        /// </summary>
        public List<Weapon> weapons;

        /// <summary>
        /// A list of armor items this bot possesses
        /// </summary>
        private List<ArmorPlate> armor;

        private int speedRating;
        private int healthRating;
        private int agilityRating;
        private int smallnessRating;

        private int aggroRating;
        private int jinkRating;
        private int freqRating;
        private int ampRating;

        private float maxSpeed { get { return speedRating * 3.5f; } }
        private float maxHealth { get { return healthRating / 2.4f + 50; } }
        private float acceleration { get { return agilityRating * 1.2f; } }
        public int radius { get { return (int)(100 / (float)Math.Log(smallnessRating, 8)); } }

        private int TargetRadius { get { return (int)(aggroRating * 2.5f + 100); } }
        private TimeSpan TotalJinkTime { get { return TimeSpan.FromSeconds(200 / jinkRating + 0.8f); } }
        private int Amplitude { get { return ampRating; } }
        private TimeSpan Period { get { return TimeSpan.FromSeconds(80f / freqRating); } }

        private float currentHealth;

        public bool Die { get { return currentHealth < 0; } }

        public Vector2 position;

        private Vector2 velocity;

        #region Success

        private TimeSpan lifeTime;
        private float damageDealt;

        /// <summary>
        /// Get a measure of how successful this bot was
        /// </summary>
        public int GetSuccess
        {
            get
            {
                return (int)(lifeTime.TotalMilliseconds + damageDealt * 700);
            }
        }

        #endregion

        public Color color;

        public Color DrawColor
        {
            get
            {
                //In spawn mode, first fades in, then solidifies from white
                float white = (float)Math.Min(1, (5000 - GameState.currentStateLength.TotalMilliseconds) / 2500f);
                if (GameState.CurrentState != State.SPAWN) white = 0;

                //Darken outside of circle
                float dist = Math.Min(512, Vector2.Distance(position, Game1.fieldCenter));
                float r = 1 - (dist / 512f) * (dist / 512f);

                //TODO: dark messes up color balance
                return new Color(
                    Math.Min(r, Math.Max(color.R / 255f, white)),
                    Math.Min(r, Math.Max(color.G / 255f, white)),
                    Math.Min(r, Math.Max(color.B / 255f, white)),
                    (GameState.CurrentState == State.SPAWN
                    ? (float)Math.Min(1, GameState.currentStateLength.TotalMilliseconds / 2500f)
                    : 1f));
            }
        }

        public Genome Genome { get { return genome; } }

        /// <summary>
        /// Create a new, empty bot
        /// </summary>
        public Bot()
        {
            weapons = new List<Weapon>();
            armor = new List<ArmorPlate>();
            AddBot(this);
        }

        /// <summary>
        /// Spawn this bot with fresh DNA
        /// </summary>
        public Bot Spawn()
        {
            genome = new Genome().Spawn();
            Build();
            return this;
        }

        /// <summary>
        /// Spawn this bot with the given DNA
        /// </summary>
        public Bot Spawn(Genome genome)
        {
            this.genome = genome;
            Build();
            return this;
        }

        /// <summary>
        /// Build this bot's DNA into physical manifestations
        /// </summary>
        public Bot Build()
        {
            speedRating = 15;
            healthRating = 15;
            agilityRating = 15;
            smallnessRating = 15;

            foreach (Atom a in genome.atoms)
            {
                switch (a.type)
                {
                    case AtomType.SPEED:
                        speedRating += a.values["value"];
                        break;
                    case AtomType.HEALTH:
                        healthRating += a.values["value"];
                        break;
                    case AtomType.AGILITY:
                        agilityRating += a.values["value"];
                        break;
                    case AtomType.SMALLNESS:
                        smallnessRating += a.values["value"];
                        break;
                    case AtomType.ARMOR:
                        armor.Add(new ArmorPlate(this, a.values));
                        break;
                    case AtomType.GUN:
                        weapons.Add(new Gun(this, a.values));
                        break;
                    case AtomType.LASER:
                        weapons.Add(new Laser(this, a.values));
                        break;
                    case AtomType.ROCKET:
                        weapons.Add(new RocketLauncher(this, a.values));
                        break;
                }
            }

            foreach (Atom a in genome.mandatoryAtoms)
            {
                switch (a.type)
                {
                    case AtomType.COLOR:
                        color = new Color(
                            1 - a.values["r"] / 100f,
                            1 - a.values["g"] / 100f,
                            1 - a.values["b"] / 100f,
                            1f);
                        break;
                    case AtomType.AGGRO:
                        aggroRating = a.values["value"];
                        break;
                    case AtomType.JINK:
                        jinkRating = a.values["value"];
                        break;
                    case AtomType.FREQ:
                        freqRating = a.values["value"];
                        break;
                    case AtomType.AMP:
                        ampRating = a.values["value"];
                        break;
                }
            }

            position = Game1.botSpawnPoint - new Vector2(0, radius);

            currentHealth = maxHealth;
            currentDistOffset = Amplitude;

            GenerateName();

            return this;
        }

        private void GenerateName()
        {
            List<string> candidateRoots = new List<string>();

            candidateRoots.Add(" sky");
            candidateRoots.Add("warrior ");
            candidateRoots.Add("agent ");
            candidateRoots.Add("orb");
            candidateRoots.Add("sphere");
            candidateRoots.Add("bot ");
            candidateRoots.Add("frame ");

            //Stat keys
            if (speedRating > 85)
            {
                candidateRoots.Add("celer");
                candidateRoots.Add(" light ");
            }
            if (speedRating > 100)
            {
                candidateRoots.Add("hermes ");
            }

            if (healthRating > 200)
            {
                candidateRoots.Add(" heavy ");
            }
            else if (healthRating > 100)
            {
                candidateRoots.Add("zeus ");
                candidateRoots.Add("bull ");
            }
            else if (healthRating > 85)
            {
                candidateRoots.Add("magnus");
            }

            if (agilityRating < 20)
            {
                candidateRoots.Add("moose ");
            }
            if (agilityRating > 100)
            {
                candidateRoots.Add("hawk ");
                candidateRoots.Add("striker ");
            }
            else if (agilityRating > 85)
            {
                candidateRoots.Add("volo");
            }

            if (smallnessRating > 200)
            {
                candidateRoots.Add("kitten ");
            }

            if (smallnessRating < 14)
            {
                candidateRoots.Add(" maximus ");
            }
            else if (smallnessRating < 25)
            {
                candidateRoots.Add("amplus");
            }
            else if (smallnessRating < 40)
            {
                candidateRoots.Add("tantus");
            }

            if (speedRating > 85 && agilityRating < 30)
            {
                candidateRoots.Add(" charging ");
            }

            //Color keys
            if (color.R < 40 && color.G < 40 && color.B < 40)
                candidateRoots.Add(" black");
            else
            {
                if (color.R > color.G && color.R > color.B)
                {
                    if (color.G - color.B > 30)
                        candidateRoots.Add(" yellow");
                    else if (color.G - color.B < -30)
                    {
                        if (color.R > 220)
                            candidateRoots.Add(" pink");
                        else
                            candidateRoots.Add(" purple");
                    }
                    else
                    {
                        if (color.R < 180 && color.R > 40)
                            candidateRoots.Add(" crimson");
                        else
                            candidateRoots.Add(" red");
                    }
                }
                if (color.G > color.R && color.G > color.B)
                {
                    if (color.R - color.B > 30)
                        candidateRoots.Add(" yellow");
                    else if (color.R - color.B < -30)
                        candidateRoots.Add(" cyan");
                    else
                        candidateRoots.Add(" green");
                }
                if (color.B > color.R && color.B > color.G)
                {
                    if (color.R - color.G > 30)
                    {
                        if (color.B > 220)
                            candidateRoots.Add(" pink");
                        else
                            candidateRoots.Add(" purple");
                    }
                    else if (color.R - color.G < -30)
                        candidateRoots.Add(" cyan");
                    else
                        candidateRoots.Add(" blue");
                }
            }

            //Weapon keys
            int laser = 0;
            int rocket = 0;
            int gun = 0;
            foreach (Weapon w in weapons)
            {
                if (w is Laser) laser++;
                if (w is RocketLauncher) rocket++;
                if (w is Gun) gun++;
            }

            if (laser > 4)
                candidateRoots.Add("panic");
            if (laser > 3)
                candidateRoots.Add("disco");
            else if (laser > 2)
                candidateRoots.Add("fractal");
            else if (laser > 1)
                candidateRoots.Add(" scintillating ");

            if (rocket > 3)
                candidateRoots.Add("titan ");
            else if (rocket > 2)
                candidateRoots.Add("saber ");
            else if (rocket > 1)
                candidateRoots.Add("rocket");

            if (gun > 3)
                candidateRoots.Add("hell");
            else if (gun > 2)
                candidateRoots.Add("scatter");
            else if (gun > 1)
                candidateRoots.Add("bullet");

            if (gun + rocket + laser > 8)
                candidateRoots.Add(" armageddon ");
            else if (gun + rocket + laser > 6)
                candidateRoots.Add(" fortress ");
            if (gun + rocket + laser < 2)
                candidateRoots.Add(" harmless ");

            //Select two or 3 keys
            string name = "";
            for (int c = 0; c < Game1.random.Next(2) + 2; c++)
            {
                int d = Game1.random.Next(candidateRoots.Count);
                if (d < candidateRoots.Count)
                {
                    name += candidateRoots[d];
                    candidateRoots.RemoveAt(d);
                }

                //Maybe add a space
                if (Game1.random.Next(8) == 0) name += " ";
            }

            //Add mark based on length of genome
            if (genome.Length > 10)
            {
                name += " Mk. " + (genome.Length - 10).ToString();
            }

            //Clean up caps and spaces
            bool lastspace = false;
            string tstring = "";
            for (int c = 0; c < name.Length; c++)
            {
                if (lastspace || c == 0)
                {
                    lastspace = false;
                    if (name[c] >= 97 && name[c] <= 122)
                        tstring += (char)(name[c] - 32);
                    else if (name[c] != ' ')
                        tstring += name[c];
                    else
                        lastspace = true;
                }
                else if (name[c] == ' ')
                {
                    lastspace = true;
                    tstring += name[c];
                }
                else
                    tstring += name[c];
            }
            genome.name = tstring;
        }

        #region Movement
        private int currentCircleDir = 1;
        private TimeSpan timeToJink;

        private int currentDistOffset;
        private TimeSpan timeToAmp;

        private double orbitPosition;

        private Vector2 targetPos;
        private float FinalTargetRadius { get { return TargetRadius + currentDistOffset; } }

        #endregion

        public void Update(GameTime gameTime)
        {
            if (GameState.CurrentState != State.BATTLE || Die) return;

            lifeTime += gameTime.ElapsedGameTime;

            /* MOVEMENT
             * Bots try to move in a circle around the player
             * They vary their distance (AGGRO root) as a sine wave (with FREQ and AMP)
             */
            orbitPosition = Math.Atan2(
                Game1.player.position.Y - position.Y,
                Game1.player.position.X - position.X) - (float)Math.PI
                + currentCircleDir * 0.1f;

            timeToJink -= gameTime.ElapsedGameTime;
            if (timeToJink <= TimeSpan.Zero)
            {
                currentCircleDir *= -1;
                timeToJink = TotalJinkTime;
            }

            timeToAmp -= gameTime.ElapsedGameTime;
            if (timeToAmp <= TimeSpan.Zero)
            {
                currentDistOffset *= -1;
                timeToAmp = Period;
            }

            targetPos = new Vector2(
                (float)(Math.Cos(orbitPosition) * FinalTargetRadius),
                (float)(Math.Sin(orbitPosition) * FinalTargetRadius))
                + Game1.player.position;

            //Move toward target
            velocity += acceleration * Vector2.Normalize(targetPos - position);

            //Limit speed
            if (velocity.Length() > maxSpeed)
                velocity = Vector2.Normalize(velocity) * maxSpeed;

            //Move
            position += velocity * (float)gameTime.ElapsedGameTime.TotalSeconds;

            //Out of bounds
            if (position.X < Game1.map.Left) position.X = Game1.map.Left;
            if (position.Y < Game1.map.Top) position.Y = Game1.map.Top;
            if (position.X > Game1.map.Right) position.X = Game1.map.Right;
            if (position.Y > Game1.map.Bottom) position.Y = Game1.map.Bottom;

            //Fire
            if (!Game1.player.Die)
                foreach (Weapon w in weapons) w.Update(gameTime);
        }

        public void Draw(SpriteBatch spriteBatch)
        {
            //Draw self
            spriteBatch.Draw(
                frame,
                Game1.TransformToDraw(position),
                null,
                DrawColor,
                0,
                new Vector2(frame.Width, frame.Height) / 2,
                radius * 2 / 100f,
                SpriteEffects.None,
                1f);

            //Draw guns
            foreach (Weapon w in weapons) w.Draw(spriteBatch);
        }

        public void DrawUI(SpriteBatch spriteBatch)
        {
            //Draw health
            spriteBatch.Draw(
                Game1.pixel,
                new Rectangle(
                    512 - (int)currentHealth, 140, (int)(currentHealth * 2), 20),
                Color.Red);

            spriteBatch.DrawString(
                Game1.readoutFont,
                genome.name,
                new Vector2(512, 80) - Game1.readoutFont.MeasureString(genome.name) / 2,
                Color.White);
        }

        public void Damage(float damage)
        {
            if (GameState.CurrentState == State.BATTLE) currentHealth -= damage;

            if (Die)
            {
                //Sound
                Game1.soundbank.PlayCue("BotDie");

                //Deactivate weapons
                foreach (Weapon w in weapons) w.Kill();

                //Fragments
                for (int c = 0; c < radius / 10; c++)
                    new Particle(
                        ParticleType.FRAG1 + Game1.random.Next(3),
                        position,
                        new Vector2(Game1.random.Next(400) - 200, -Game1.random.Next(300) - 200),
                        new Vector2(0, 250),
                        TimeSpan.FromSeconds(1.5),
                        (float)Game1.random.NextDouble() * 6,
                        color,
                        ParticleType.SMOKE,
                        TimeSpan.FromMilliseconds(60),
                        radius * 2 / 100f);

                //Fire
                for (int c = 0; c < 8; c++)
                {
                    Vector2 velocity = new Vector2(
                        Game1.random.Next(100) - 50,
                        -Game1.random.Next(300));
                    new Particle(
                        ParticleType.FLAME, position,
                        velocity,
                        new Vector2(0, 0),
                        TimeSpan.FromMilliseconds(500),
                        false);
                    new Particle(
                        ParticleType.SMOKE, position,
                        velocity,
                        new Vector2(0, 0),
                        TimeSpan.FromMilliseconds(1000),
                        true);
                }
            }
        }

        public void DealtDamage(float damage)
        {
            damageDealt += damage;
        }

        public int CompareTo(Object obj)
        {
            //Sort order inverted - higher comes first
            return ((Bot)obj).GetSuccess.CompareTo(GetSuccess);
        }
    }
}
