﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Audio;

namespace Primeval
{
    /// <summary>
    /// Represents a weapon that can be stuck onto a bot
    /// </summary>
    abstract class Weapon
    {
        protected Bot owner;

        public Weapon(Bot owner)
        {
            this.owner = owner;
        }

        public abstract void Update(GameTime gameTime);

        public abstract void Draw(SpriteBatch spriteBatch);

        public virtual void Kill()
        {

        }
    }


    class Gun : Weapon
    {
        private int spreadRating;
        private int fspeedRating;
        private int pspeedRating;
        private int damageRating;
        private int clipRating;
        private int reloadRating;

        public int SpreadCount { get { return spreadRating / 20; } }
        public float SpreadDist { get { return (spreadRating % 20) / 30f; } }
        public TimeSpan Cooldown { get { return TimeSpan.FromSeconds(0.6f / (float)Math.Log(fspeedRating, 8)); } }
        public float ProjectileVelocity { get { return 350 * (float)Math.Log(pspeedRating, 8); } }
        public float ProjectileDamage { get { return 2 * (float)Math.Log(damageRating, 8); } }
        public int ClipSize { get { return (int)Math.Ceiling(clipRating / 11f); } }
        public TimeSpan ReloadTime { get { return TimeSpan.FromSeconds(5f / (float)Math.Log(fspeedRating, 8)); } }
        public float Radius { get { return (float)Math.Log(ProjectileDamage, 9) * 16; } }

        private TimeSpan currentCooldown;
        private TimeSpan currentReloadTime;
        private int currentClip;

        public Gun(Bot owner, Dictionary<string, AtomValue> values)
            : base(owner)
        {
            spreadRating = values["spread"];
            fspeedRating = values["fspeed"];
            pspeedRating = values["pspeed"];
            damageRating = values["damage"];
            clipRating = values["clipsize"];
            reloadRating = values["reloadspeed"];

            currentClip = ClipSize;
        }

        public override void Update(GameTime gameTime)
        {
            currentCooldown -= gameTime.ElapsedGameTime;
            currentReloadTime -= gameTime.ElapsedGameTime;
            if (currentCooldown <= TimeSpan.Zero
                && currentReloadTime <= TimeSpan.Zero)
            {
                //Fire
                float rootang = (float)Math.Atan2(
                    Game1.player.position.Y - owner.position.Y,
                    Game1.player.position.X - owner.position.X);
                int step = 0;
                for (int c = 0; c < SpreadCount; c++)
                {
                    new Bullet(ProjectileVelocity, ProjectileDamage, Radius,
                        new Vector2(
                            (float)Math.Cos(rootang + step * SpreadDist),
                            (float)Math.Sin(rootang + step * SpreadDist)),
                            owner);

                    if (step == 0)
                        step = -1;
                    else if (step < 0)
                        step = -step;
                    else
                        step = -step - 1;
                }

                currentClip--;
                if (currentClip <= 0)
                {
                    currentClip = ClipSize;
                    currentReloadTime = ReloadTime;
                }
                currentCooldown = Cooldown;
            }
        }

        public override void Draw(SpriteBatch spriteBatch)
        {
            
        }
    }


    class Laser : Weapon
    {
        private int durationRating;
        private int cooldownRating;
        private int damageRating;
        private int arcstartRating;
        private int arcendRating;

        public TimeSpan Duration { get { return TimeSpan.FromSeconds(0.8f * (float)Math.Log(durationRating, 8)); } }
        public TimeSpan Cooldown { get { return TimeSpan.FromSeconds(5.5f / (float)Math.Log(cooldownRating, 8)); } }
        public float DPS { get { return damageRating / 3f; } }
        public float ArcStart { get { return arcstartRating / 100f * (float)Math.PI; } }
        public float ArcEnd { get { return arcendRating / 100f * (float)Math.PI; } }

        public double Angle { get { return (ArcStart - ArcEnd) * currentDuration.TotalSeconds + ArcStart; } }
        public Vector2 Firedir
        {
            get
            {
                return Vector2.Normalize(
                    new Vector2((float)Math.Cos(Angle), (float)Math.Sin(Angle)));
            }
        }

        public Color color;

        private TimeSpan currentCooldown;
        private TimeSpan currentDuration;

        private static Texture2D beamimage;

        private Cue cue;

        public float radius { get { return DPS / 1.5f; } }

        public static void LoadContent()
        {
            beamimage = Game1.content.Load<Texture2D>("laser");
        }

        public Laser(Bot owner, Dictionary<string, AtomValue> values)
            : base(owner)
        {
            durationRating = values["duration"];
            cooldownRating = values["cooldown"];
            damageRating = values["damage"];
            arcstartRating = values["arcstart"];
            arcendRating = values["arcend"];
            color = new Color(
                1 - (values["r"] * values["r"]) / 1000f,
                1 - (values["g"] * values["g"]) / 1000f,
                1 - (values["b"] * values["b"]) / 1000f,
                1f);
        }

        public override void Update(GameTime gameTime)
        {
            currentCooldown -= gameTime.ElapsedGameTime;
            if (currentCooldown <= TimeSpan.Zero)
            {
                if (currentDuration == TimeSpan.Zero)
                    currentDuration = Duration;
                else
                {
                    if (cue == null)
                    {
                        cue = Game1.soundbank.GetCue("Laser");
                        cue.Play();
                    }
                    currentDuration -= gameTime.ElapsedGameTime;
                }

                //Damage
                float distToPlayer = ((owner.position - Game1.player.position)
                    - Vector2.Dot(owner.position - Game1.player.position, Firedir) * Firedir).Length();
                if (distToPlayer <= Player.radius + radius
                    && Math.Sign(Firedir.Y) == Math.Sign(Game1.player.position.Y - owner.position.Y)
                    && Math.Sign(Firedir.X) == Math.Sign(Game1.player.position.X - owner.position.X))
                {
                    Game1.player.Damage(DPS * (float)gameTime.ElapsedGameTime.TotalSeconds);
                    owner.DealtDamage(DPS * (float)gameTime.ElapsedGameTime.TotalSeconds);

                    //Spawn particles
                    Vector2 normal = Vector2.Normalize(owner.position - Game1.player.position);
                    if (Game1.random.Next(3) == 0)
                        new Particle(
                            ParticleType.SPARK,
                            Game1.player.position + normal * (Player.radius + 15),
                            normal * (float)Game1.random.NextDouble() * 300
                            + new Vector2((float)Game1.random.NextDouble() * 60, (float)Game1.random.NextDouble() * 60),
                            Vector2.Zero,
                            TimeSpan.FromMilliseconds(550),
                            0,
                            color);
                }

                if (currentDuration <= TimeSpan.Zero)
                {
                    currentDuration = TimeSpan.Zero;
                    currentCooldown = Cooldown;
                    cue.Stop(AudioStopOptions.AsAuthored);
                    cue.Dispose();
                    cue = null;
                }
            }
        }

        public override void Kill()
        {
            if (cue != null)
            {
                cue.Stop(AudioStopOptions.AsAuthored);
                cue.Dispose();
                cue = null;
            }
        }

        public override void Draw(SpriteBatch spriteBatch)
        {
            if (currentDuration > TimeSpan.Zero)
            {
                Vector2 tpos = Game1.TransformToDraw(new Vector2(
                    (float)Math.Cos(Angle),
                    (float)Math.Sin(Angle)) * 2000 + owner.position);
                Vector2 spos = Game1.TransformToDraw(owner.position);

                Vector2 dir = new Vector2(tpos.X - spos.X, tpos.Y - spos.Y);
                float totallen = dir.Length();
                dir.Normalize();

                float angtoplay = (float)Math.Atan2(dir.Y, dir.X) - (float)Math.PI / 2;

                //Grow
                float pos = (float)Math.Min(1,
                    Math.Min(
                        currentDuration.TotalMilliseconds,
                        (Duration - currentDuration).TotalMilliseconds) / 100);

                for (int dist = 0; dist < totallen; dist += beamimage.Height - 2)
                {
                    spriteBatch.Draw(
                        beamimage,
                        new Rectangle(
                            (int)(spos.X + dir.X * dist),
                            (int)(spos.Y + dir.Y * dist),
                            (int)(radius * 2 * pos),
                            (int)Math.Min(totallen - dist, beamimage.Height)),
                        null,
                        color,
                        angtoplay,
                        new Vector2(beamimage.Width / 2, 0),
                        SpriteEffects.None,
                        1f);
                }
            }
        }
    }


    class RocketLauncher : Weapon
    {
        private int fspeedRating;
        private int maxspeedRating;
        private int accelerationRating;
        private int damageRating;
        private int directionRating;
        private int pollutionRating;

        public TimeSpan Cooldown { get { return TimeSpan.FromSeconds(5f / (float)Math.Log(fspeedRating, 8)); } }
        public float RocketVelocity { get { return 300 * (float)Math.Log(maxspeedRating, 8); } }
        public float RocketAccel { get { return 600 * (float)Math.Log(maxspeedRating, 8); } }
        public float RocketDamage { get { return 8 * (float)Math.Log(damageRating, 8); } }
        public double RocketDirection { get { return Math.PI * directionRating / 50f; } }
        public int PollutionInterval { get { return (int)(pollutionRating / 2f + 20); } }
        public float Radius { get { return 12 * (float)Math.Log(RocketDamage, 7); } }

        private TimeSpan currentCooldown;

        public RocketLauncher(Bot owner, Dictionary<string, AtomValue> values)
            : base(owner)
        {
            fspeedRating = values["fspeed"];
            maxspeedRating = values["maxspeed"];
            accelerationRating = values["agility"];
            damageRating = values["damage"];
            directionRating = values["direction"];
            pollutionRating = values["pollution"];
        }

        public override void Update(GameTime gameTime)
        {
            currentCooldown -= gameTime.ElapsedGameTime;
            if (currentCooldown <= TimeSpan.Zero)
            {
                new Rocket(RocketVelocity, RocketDamage, Radius,
                    new Vector2(
                        (float)Math.Cos(RocketDirection),
                        (float)Math.Sin(RocketDirection)),
                    owner, RocketAccel, PollutionInterval);

                currentCooldown = Cooldown;
            }
        }

        public override void Draw(SpriteBatch spriteBatch)
        {

        }
    }
}
